// Copyright (c) Microsoft Corporation. All rights reserved.
// Copyright (c) 2023 SpacemiT. All rights reserved.
// Licensed under the MIT License.

#include "ort_genai.h"
#include <algorithm>
#include <chrono>
#include <cstring>
#include <iostream>
#include <memory>
#include <string>
#include <sys/syscall.h>
#include <unistd.h>

// C++ API Example

std::string GetChatTemplate(const char *model_type, std::string &text) {
  if (strcmp(model_type, "phi3") == 0) {
    return "<|user|>\n" + text + "<|end|>\n<|assistant|>";
  } else if (strcmp(model_type, "qwen2") == 0) {
    return "<|im_start|>user\n" + text + "<|im_end|><|im_start|>assistant";
  } else if (strcmp(model_type, "minicpm") == 0) {
    return "<s><用户>\n" + text + "<AI>";
  } else if (strcmp(model_type, "llama3") == 0) {
    return "<|start_header_id|>user<|end_header_id|>" + text +
           "<|eot_id|><|start_header_id|>assistant<|end_header_id|>";
  } else {
    throw std::runtime_error("Unknown model type");
  }
}

void CXX_API(const char *model_path, const char *model_type) {

  // 开启日志打印
  // Oga::SetLogBool("enabled", true);
  // 开启日志打印模型输入值
  // Oga::SetLogBool("model_input_values", true);
  // 设置日志文件名
  // Oga::SetLogString("filename", "ort-genai.log");

  //std::cout << "Creating model..." << std::endl;
  auto model = OgaModel::Create(model_path);
  //std::cout << "Creating tokenizer..." << std::endl;
  auto tokenizer = OgaTokenizer::Create(*model);
  auto tokenizer_stream = OgaTokenizerStream::Create(*tokenizer);

  while (true) {
    std::string text;
    std::cout << "Prompt(enter 'stop' to exit): " << std::endl;
    std::getline(std::cin, text);

    if (text == "stop") {
      break;
    }

    const std::string prompt = GetChatTemplate(model_type, text);

    auto sequences = OgaSequences::Create();
    tokenizer->Encode(prompt.c_str(), *sequences);
    auto params = OgaGeneratorParams::Create(*model);
    params->SetSearchOption("max_length", 1024);

    //std::cout << "Generating response..." << std::endl;

    params->SetInputSequences(*sequences);
    auto generator = OgaGenerator::Create(*model, *params);

    int64_t token_generated = 0;
    float first_token_generated_time = 0.0f;
    float avg_token_generated_time = 0.0f;
    bool first_token_generated = false;
    int64_t prompt_num_tokens = sequences->SequenceCount(0);

    while (!generator->IsDone()) {
      auto time_start = std::chrono::high_resolution_clock::now();
      generator->ComputeLogits();
      generator->GenerateNextToken();

      const auto num_tokens = generator->GetSequenceCount(0);
      const auto new_token = generator->GetSequenceData(0)[num_tokens - 1];
      auto time_end = std::chrono::high_resolution_clock::now();
      if (first_token_generated) {
        avg_token_generated_time +=
            std::chrono::duration_cast<std::chrono::microseconds>(time_end -
                                                                  time_start)
                .count();
        token_generated += 1;
      } else {
        first_token_generated = true;
        first_token_generated_time =
            std::chrono::duration_cast<std::chrono::microseconds>(time_end -
                                                                  time_start)
                .count();
      }
      std::cout << tokenizer_stream->Decode(new_token) << std::flush;
    }

    //std::cout << std::endl;
    //std::cout << "first token latency: " << prompt_num_tokens
    //          << " tokens, time: " << first_token_generated_time / 1000.0f
    //          << " ms, generated " << token_generated
    //          << " tokens, time: " << avg_token_generated_time / 1000.0f
    //          << " ms, "
    //          << token_generated /
    //                 (avg_token_generated_time / 1000.0f / 1000.0f)
    //          << " tps" << std::endl;

    for (int i = 0; i < 3; ++i)
      std::cout << std::endl;
  }
}

// C API Example

void CheckResult(OgaResult *result) {
  if (result) {
    std::string string = OgaResultGetError(result);
    OgaDestroyResult(result);
    throw std::runtime_error(string);
  }
}

void C_API(const char *model_path, const char *model_type) {
  OgaModel *model;
  //std::cout << "Creating model..." << std::endl;
  OgaCreateModel(model_path, &model);

  OgaTokenizer *tokenizer;
  //std::cout << "Creating tokenizer..." << std::endl;
  CheckResult(OgaCreateTokenizer(model, &tokenizer));

  OgaTokenizerStream *tokenizer_stream;
  CheckResult(OgaCreateTokenizerStream(tokenizer, &tokenizer_stream));

  while (true) {
    std::string text;
    std::cout << "Prompt(enter 'stop' to exit): " << std::endl;
    std::getline(std::cin, text);

    if (text == "stop") {
      break;
    }

    const std::string prompt = GetChatTemplate(model_type, text);

    OgaSequences *sequences;
    CheckResult(OgaCreateSequences(&sequences));
    CheckResult(OgaTokenizerEncode(tokenizer, prompt.c_str(), sequences));

    //std::cout << "Generating response..." << std::endl;
    OgaGeneratorParams *params;
    CheckResult(OgaCreateGeneratorParams(model, &params));
    CheckResult(OgaGeneratorParamsSetSearchNumber(params, "max_length", 1024));
    CheckResult(OgaGeneratorParamsSetInputSequences(params, sequences));

    OgaGenerator *generator;
    CheckResult(OgaCreateGenerator(model, params, &generator));

    int64_t token_generated = 0;
    float first_token_generated_time = 0.0f;
    float avg_token_generated_time = 0.0f;
    bool first_token_generated = false;
    int64_t prompt_num_tokens = sequences->SequenceCount(0);

    while (!OgaGenerator_IsDone(generator)) {
      auto time_start = std::chrono::high_resolution_clock::now();

      CheckResult(OgaGenerator_ComputeLogits(generator));
      CheckResult(OgaGenerator_GenerateNextToken(generator));

      auto time_end = std::chrono::high_resolution_clock::now();
      if (first_token_generated) {
        avg_token_generated_time +=
            std::chrono::duration_cast<std::chrono::microseconds>(time_end -
                                                                  time_start)
                .count();
        token_generated += 1;
      } else {
        first_token_generated = true;
        first_token_generated_time =
            std::chrono::duration_cast<std::chrono::microseconds>(time_end -
                                                                  time_start)
                .count();
      }

      const int32_t num_tokens = OgaGenerator_GetSequenceCount(generator, 0);
      int32_t new_token =
          OgaGenerator_GetSequenceData(generator, 0)[num_tokens - 1];
      const char *new_token_string;
      CheckResult(OgaTokenizerStreamDecode(tokenizer_stream, new_token,
                                           &new_token_string));
      std::cout << new_token_string << std::flush;
    }

    //std::cout << std::endl;
    //std::cout << "first token latency: " << prompt_num_tokens
    //          << " tokens, time: " << first_token_generated_time / 1000.0f
    //          << " ms, generated " << token_generated
    //          << " tokens, time: " << avg_token_generated_time / 1000.0f
    //          << " ms, "
    //          << token_generated /
    //                 (avg_token_generated_time / 1000.0f / 1000.0f)
    //          << " tps" << std::endl;

    for (int i = 0; i < 3; ++i)
      std::cout << std::endl;

    OgaDestroyGeneratorParams(params);
    OgaDestroySequences(sequences);
  }

  OgaDestroyTokenizerStream(tokenizer_stream);
  OgaDestroyTokenizer(tokenizer);
  OgaDestroyModel(model);
}

static void print_usage(int /*argc*/, char **argv) {
  std::cerr << "usage: " << argv[0] << " model_path  model_type[phi3|qwen2|minicpm|llama3]" << std::endl;
}

int main(int argc, char **argv) {
  if (argc != 3) {
    print_usage(argc, argv);
    return -1;
  }

  // Responsible for cleaning up the library during shutdown
  OgaHandle handle;

  //std::cout << "-------------" << std::endl;
  std::cout << "Hello, " << argv[2] << "!" << std::endl;
  //std::cout << "-------------" << std::endl;

#ifdef USE_CXX
  //std::cout << "C++ API" << std::endl;
  CXX_API(argv[1], argv[2]);
#else
  //std::cout << "C API" << std::endl;
  C_API(argv[1], argv[2]);
#endif

  return 0;
}
